/*

DESCRIPTION:
    Main source file for 12VDC->5VDC SMPS non-isolated buck based on full-bridge
    inverter development using PIC16F1769.

SYNTAX:  n/a

KNOWN BUGS:  n/a

COPYRIGHT:  (c) 2017, Brian Cornell, all rights reserved.

DEPENDENCIES:
    main.h - configuration constants for main.c

VERSION:  1.1, production release

MODIFICATION HISTORY:

    28 Oct 17, BWC, created file.
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "config.h"
#include "pval.h"
#include "main.h"

//#define _TESTPWM                        //Uncomment to use test pattern


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Data objects
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
unsigned char PWMindex;         //Index to active element in array PWMTable
bit sineDir,                    //Sine angle increasing/decreasing
#ifdef _CFGFB
        sinePol,                //Sine polarity
#endif
        statRun,                //Bridge status - run/stop
        cmdStop;                //Commands ISR to finish current halversine & stop

//Production pattern for 20kHz PWM
#ifndef _TESTPWM
const uint16_t PWMtable[S_QY_PWMQSCYC] = {P20V0,P20V1,P20V2,P20V3,P20V4,P20V5,
    P20V6,P20V7,P20V8,P20V9,P20V10,P20V11,P20V12,P20V13,P20V14,P20V15,P20V16,P20V17,
    P20V18,P20V19,P20V20,P20V21,P20V22,P20V23,P20V24,P20V25,P20V26,P20V27,P20V28,
    P20V29,P20V30,P20V31,P20V32,P20V33,P20V34,P20V35,P20V36,P20V37,P20V38,P20V39,
    P20V40,P20V41,P20V42,P20V43,P20V44,P20V45,P20V46,P20V47,P20V48,P20V49,P20V50,
    P20V51,P20V52,P20V53,P20V54,P20V55,P20V56,P20V57,P20V58,P20V59,P20V60,P20V61,
    P20V62,P20V63,P20V64,P20V65,P20V66,P20V67,P20V68,P20V69,P20V70,P20V71,P20V72,
    P20V73,P20V74,P20V75,P20V76,P20V77,P20V78,P20V79,P20V80,P20V81,P20V82, P20V83,
    P20V84,P20V85,P20V86,P20V87,P20V88,P20V89,P20V90,P20V91,P20V92,P20V93,P20V94,
    P20V95,P20V96,P20V97,P20V98,P20V99};
#else
//Test pattern (constant) for 20kHz PWM
const uint16_t PWMtable[S_QY_PWMQSCYC] = {P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,P20TP,
    P20TP,P20TP};
#endif



/*Function main.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    uint8_t count;

    //Configure system and pause settle time
    Initialize();
    __delay_ms(S_TM_INT);

    while (TRUE)
    {
        if ((PIN_I_RUN == S_CD_RUN) && (statRun == S_CD_STOP))
        {
            //Start bridge
            PWMindex = 0;
            sineDir = S_BL_INCR;
#ifdef _CFGFB
            sinePol = S_BL_POS;
#endif
#ifdef _CFGHB
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set bridge enable pins to start with A high-side active.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            PIN_O_ENA = TRUE;
            PIN_O_ENB = FALSE;
#endif

            TMR2 = 0;                       //Clear timer count
            TMR2IF = FALSE;                 //Clear timer interrupt flag
            PWM5DCH = (uint8_t) (PWMtable[PWMindex] >> CFG_PWM_DCH_BITS);
            PWM5DCL = (uint8_t) ((PWMtable[PWMindex] & CFG_PWM_DCL_MASK) << CFG_PWM_DCL_BITS);
            PWM5EN = TRUE;
            TMR2ON = TRUE;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Let timer2 run for S_LM_TMR2START periods before starting CWG to
             * ensure a clean start (0 angle) for sine wave.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            for (count = 0; (count < S_LM_TMR2START); )
            {
                if (TMR2IF)
                {
                    ++count;
                    TMR2IF = FALSE;
                }
            }
            CWG1AS0bits.SHUTDOWN = FALSE;
            TMR2IE = TRUE;
            statRun = S_CD_RUN;
        }

        if ((PIN_I_RUN == S_CD_STOP) && (statRun == S_CD_RUN))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Stop bridge:  issue command & wait for ISR to clear.  Then 
             * shutdown CWG & supporting peripherals.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmdStop = TRUE;
            while (cmdStop);
            CWG1AS0bits.SHUTDOWN = TRUE;
#ifdef _CFGHB
            PIN_O_ENA = FALSE;
            PIN_O_ENB = FALSE;
#endif
            TMR2IE = FALSE;
            PWM5EN = FALSE;
            TMR2ON = FALSE;
            statRun = S_CD_STOP;
        }
    }
} //End function main


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts:  timer2.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    static unsigned int dc;

    if (TMR2IF)
    {
        //Fetch next DC
        dc = PWMtable[PWMindex];
        //Load next duty cycle value
        PWM5DCH = (uint8_t) (dc >> CFG_PWM_DCH_BITS);
        PWM5DCL = (uint8_t) ((dc & CFG_PWM_DCL_MASK) << CFG_PWM_DCL_BITS);

        //Check direction & if at peak reverse
        if (sineDir == S_BL_INCR)
        {
            if (PWMindex < S_LM_PWMINDEX) ++PWMindex;
            else sineDir = S_BL_DECR;
        }
        else
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * When index reaches zero reverse polarity and begin next 
             * halversine.  If stop bit set then clear interrupt enable to
             * prevent further generation.  Foreground process is expected to
             * stop CWG.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (PWMindex > 0) --PWMindex;
            else
            {
                if (cmdStop)
                {
                    TMR2IE = FALSE;
                    cmdStop = FALSE;
                }
                else
                {
                    //Reverse polarity every 180 degrees
                    sineDir = S_BL_INCR;
#ifdef _CFGFB
                    sinePol ^= S_BL_NEG;
                    CWG1CON0bits.MODE0 = sinePol;
#endif
#ifdef _CFGHB
                    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                     * The high side output must be off (e.g. duty cycle 
                     * completed) before direction bits can be flipped to 
                     * prevent shoot-thru.  To Execution is suspended until the 
                     * PWM output is FALSE.  Since the DC at 0 or 180 degrees is 
                     * nearly zero the wait time will be very small (<1uS).
                     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                    while (S_BL_HSON);
                    PIN_O_ENA ^= TRUE;
                    PIN_O_ENB ^= TRUE;
#endif
                }
            }
        }

        TMR2IF = FALSE;
    }

    return;
} //End function InterruptExec


/*Function Initialize.  Performs baseline MPU initialization at boot.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure system clock.  Expects 1MHz HFINTOSC to be active at boot.
     * Adjust HF internal oscillator to obtain 60Hz period from TMR2.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON1bits.NOSC = CFG_OSC_SRC;
    OSCCON1bits.NDIV = CFG_OSC_DIV;
    OSCFRQ = CFG_OSC_FRQ;
    while (!(OSCSTAT1bits.HFOR));       //Make sure oscillator is online
    OSCTUNE = CFG_OSC_OSCTUNE;          //Shift oscillator frequency if necessary


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important but miscellaneous settings that must be set at boot and aren't
     * part of a larger group.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _CFGFB
    sinePol = S_BL_POS;
#endif
    sineDir = S_BL_INCR;
    statRun = S_CD_STOP;
    cmdStop = FALSE;
    PWMindex = 0;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set port configuration.  To prevent any aberrant behaviors before 
     * initialization is complete set output latches.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //Port A
    PORTA = VL_PORTA;
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port C
    PORTC = VL_PORTC;
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Map peripheral pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _CFGFB
    //CWG & CLC outputs
    PPS_P_CWG1_A = PPS_M_CWG1_A;
    PPS_P_CWG1_D = PPS_M_CWG1_D;
    PPS_P_CWG1_C = PPS_M_CWG1_C;
    PPS_P_CWG1_B = PPS_M_CWG1_B;
    PPS_P_CLC1 = PPS_M_CLC1;
    PPS_P_CLC2 = PPS_M_CLC2;

    //CLC inputs
    CLCIN0PPS = PPS_M_CLC_0;
    CLCIN1PPS = PPS_M_CLC_1;
    CLCIN2PPS = PPS_M_CLC_2;
    CLCIN3PPS = PPS_M_CLC_3;
#endif

#ifdef _CFGHB
    //CWG & CLC outputs
    PPS_P_CWG1_A = PPS_M_CWG1_A;
    PPS_P_CWG1_B = PPS_M_CWG1_B;
    PPS_P_CLC1 = PPS_M_CLC1;
    PPS_P_CLC2 = PPS_M_CLC2;
    PPS_P_CLC3 = PPS_M_CLC3;
    PPS_P_CLC4 = PPS_M_CLC4;

    //CLC inputs
    CLCIN0PPS = PPS_M_CLC_0;
    CLCIN1PPS = PPS_M_CLC_1;
    CLCIN2PPS = PPS_M_CLC_2;
    CLCIN3PPS = PPS_M_CLC_3;
#endif


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure Timer2
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T2CON = CFG_TMR_2_CON;
    PR2 = CFG_TMR_2_FRQ;
    
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM5
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5CON = CFG_PWM_5_CON;
    PWMTMRS = CFG_PWM_5_TMRS;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CWG1.  During config CWG should be disabled.  Once done auto-
     * shutdown bit should be set to prevent spurious operation so unit can be 
     * enabled.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CWG1CON0 = CFG_CWG_1_CON0;
    CWG1CON1 = CFG_CWG_1_CON1;
    CWG1CLKCON = CFG_CWG_1_CLK;
    CWG1DAT = CFG_CWG_1_DAT;
    CWG1AS0 = CFG_CWG_1_AS0;
    CWG1AS1 = CFG_CWG_1_AS1;
    CWG1DBR = CFG_CWG_1_DBR;
    CWG1DBF = CFG_CWG_1_DBF;
    CWG1CON0bits.EN = TRUE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLCs.  Once configured enable.  This ensures output remains low 
     * (dependent on CWG being configured and placed in AS state).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //CLC1
    CLC1CON = CFG_CLC_1_CON;
    CLC1POL = CFG_CLC_1_POL;
    CLC1SEL0 = CFG_CLC_1_SEL0;
    CLC1SEL1 = CFG_CLC_1_SEL1;
    CLC1SEL2 = CFG_CLC_1_SEL2;
    CLC1SEL3 = CFG_CLC_1_SEL3;
    CLC1GLS0 = CFG_CLC_1_LS_0;
    CLC1GLS1 = CFG_CLC_1_LS_1;
    CLC1GLS2 = CFG_CLC_1_LS_2;
    CLC1GLS3 = CFG_CLC_1_LS_3;
    CLC1CONbits.EN = TRUE;

    //CLC2
    CLC2CON = CFG_CLC_2_CON;
    CLC2POL = CFG_CLC_2_POL;
    CLC2SEL0 = CFG_CLC_2_SEL0;
    CLC2SEL1 = CFG_CLC_2_SEL1;
    CLC2SEL2 = CFG_CLC_2_SEL2;
    CLC2SEL3 = CFG_CLC_2_SEL3;
    CLC2GLS0 = CFG_CLC_2_LS_0;
    CLC2GLS1 = CFG_CLC_2_LS_1;
    CLC2GLS2 = CFG_CLC_2_LS_2;
    CLC2GLS3 = CFG_CLC_2_LS_3;
    CLC2CONbits.EN = TRUE;

#ifdef _CFGHB
    //CLC3
    CLC3CON = CFG_CLC_3_CON;
    CLC3POL = CFG_CLC_3_POL;
    CLC3SEL0 = CFG_CLC_3_SEL0;
    CLC3SEL1 = CFG_CLC_3_SEL1;
    CLC3SEL2 = CFG_CLC_3_SEL2;
    CLC3SEL3 = CFG_CLC_3_SEL3;
    CLC3GLS0 = CFG_CLC_3_LS_0;
    CLC3GLS1 = CFG_CLC_3_LS_1;
    CLC3GLS2 = CFG_CLC_3_LS_2;
    CLC3GLS3 = CFG_CLC_3_LS_3;
    CLC3CONbits.EN = TRUE;

    //CLC4
    CLC4CON = CFG_CLC_4_CON;
    CLC4POL = CFG_CLC_4_POL;
    CLC4SEL0 = CFG_CLC_4_SEL0;
    CLC4SEL1 = CFG_CLC_4_SEL1;
    CLC4SEL2 = CFG_CLC_4_SEL2;
    CLC4SEL3 = CFG_CLC_4_SEL3;
    CLC4GLS0 = CFG_CLC_4_LS_0;
    CLC4GLS1 = CFG_CLC_4_LS_1;
    CLC4GLS2 = CFG_CLC_4_LS_2;
    CLC4GLS3 = CFG_CLC_4_LS_3;
    CLC4CONbits.EN = TRUE;
#endif


    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts

    return;
} //End function Initialize



//End of file